"""
Provide main entry point for the Jobbergate CLI App.
"""

import sys

import httpx
import importlib_metadata
import typer
from jobbergate_core.auth.handler import JobbergateAuthHandler

from jobbergate_cli.auth import show_login_message, track_login_progress
from jobbergate_cli.config import settings
from jobbergate_cli.exceptions import Abort, handle_abort, handle_authentication_error
from jobbergate_cli.logging import init_logs, init_sentry
from jobbergate_cli.render import render_demo, render_json, terminal_message
from jobbergate_cli.schemas import JobbergateContext
from jobbergate_cli.subapps.applications.app import app as applications_app
from jobbergate_cli.subapps.job_scripts.app import app as job_scripts_app
from jobbergate_cli.subapps.job_submissions.app import app as job_submissions_app
from jobbergate_cli.text_tools import copy_to_clipboard

app = typer.Typer()


# If "compatibility" mode is set through the environment, map the commands at their familiar placement on the main app.
if settings.JOBBERGATE_COMPATIBILITY_MODE:
    from jobbergate_cli.compat import add_legacy_compatible_commands

    add_legacy_compatible_commands(app)

app.add_typer(applications_app, name="applications")
app.add_typer(job_scripts_app, name="job-scripts")
app.add_typer(job_submissions_app, name="job-submissions")


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    verbose: bool = typer.Option(False, help="Enable verbose logging to the terminal"),
    full: bool = typer.Option(False, help="Print all fields from CRUD commands"),
    raw: bool = typer.Option(False, help="Print output from CRUD commands as raw json"),
    version: bool = typer.Option(False, help="Print the version of jobbergate-cli and exit"),
    ignore_extra_args: str = typer.Option(
        None,
        "--username",
        "-u",
        "--password",
        "-p",
        hidden=True,
        help="Ignore extra arguments passed to the command for backward compatibility with the legacy app.",
    ),
):
    """
    Welcome to the Jobbergate CLI!

    More information can be shown for each command listed below by running it with the --help option.
    """
    if version:
        typer.echo(importlib_metadata.version("jobbergate-cli"))
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        render_demo(pre_amble="No command provided.")
        raise typer.Exit()

    init_logs(verbose=verbose)
    init_sentry()

    protocol = "https" if settings.OIDC_USE_HTTPS else "http"
    authentication_handler = JobbergateAuthHandler(
        cache_directory=settings.JOBBERGATE_USER_TOKEN_DIR,
        login_domain=f"{protocol}://{settings.OIDC_DOMAIN}",
        login_audience=settings.OIDC_AUDIENCE,
        login_client_id=settings.OIDC_CLIENT_ID,
        login_client_secret=settings.OIDC_CLIENT_SECRET,
        login_url_handler=show_login_message,
        login_sequence_handler=track_login_progress,
    )

    ctx.obj = JobbergateContext(
        client=httpx.Client(
            base_url=settings.ARMADA_API_BASE,
            auth=authentication_handler,
            timeout=settings.JOBBERGATE_REQUESTS_TIMEOUT,
        ),
        authentication_handler=authentication_handler,
        full_output=full,
        raw_output=raw,
    )


@app.command(rich_help_panel="Authentication")
def login(ctx: typer.Context):
    """
    Log in to the jobbergate-cli by storing the supplied token argument in the cache.
    """
    ctx.obj.authentication_handler.login()
    identity_data = ctx.obj.authentication_handler.get_identity_data()
    terminal_message(f"User was logged in with email '{identity_data.email}'", subject="Logged in!")
    render_demo()


@app.command(rich_help_panel="Authentication")
def logout(ctx: typer.Context):
    """
    Logs out of the jobbergate-cli. Clears the saved user credentials.
    """
    ctx.obj.authentication_handler.logout()
    terminal_message("User was logged out.", subject="Logged out")


@app.command(rich_help_panel="Authentication")
def show_token(
    ctx: typer.Context,
    plain: bool = typer.Option(
        False,
        help="Show the token in plain text.",
    ),
    refresh: bool = typer.Option(
        False,
        help="Show the refresh token instead of the access token.",
    ),
    show_prefix: bool = typer.Option(
        False,
        "--prefix",
        help="Include the 'Bearer' prefix in the output.",
    ),
    show_header: bool = typer.Option(
        False,
        "--header",
        help="Show the token as it would appear in a request header.",
    ),
    decode: bool = typer.Option(
        False,
        "--decode",
        help="Show the content of the decoded access token.",
    ),
):
    """
    Show the token for the logged in user.

    Token output is automatically copied to your clipboard.
    """
    ctx.obj.authentication_handler.acquire_access()
    if refresh:
        token = ctx.obj.authentication_handler._refresh_token
    else:
        token = ctx.obj.authentication_handler._access_token

    Abort.require_condition(token.is_valid(), f"Could not obtain {token.label}. Please try loggin in again.")

    if decode:
        # Decode the token with ALL verification turned off (we just want to unpack it)
        render_json(token.data)
        return

    if show_header:
        token_text = f"""{{ "Authorization": "{token.bearer_token}" }}"""
    elif show_prefix:
        token_text = token.bearer_token
    else:
        token_text = token.content

    on_clipboard = copy_to_clipboard(token_text)

    if plain:
        print(token_text)
    else:
        subject = f"{token.label.title()} Token"
        kwargs = dict(subject=subject, indent=False)
        if on_clipboard:
            kwargs["footer"] = "The output was copied to your clipboard"

        terminal_message(token_text, **kwargs)


def safe_entrypoint():
    """
    Entrypoint for the app including custom error handling.

    With this we ensure error handling is applied to all commands with no need
    to duplicate the decorators on each of them.
    """
    try:
        safe_function = handle_abort(handle_authentication_error(app.__call__))
        safe_function()
    except typer.Exit as e:
        sys.exit(e.exit_code)



"""
Provide a ``typer`` app that can interact with Application data in a cruddy manner.
"""

import pathlib
from textwrap import dedent
from typing import Any, Dict, Optional, cast

import typer

from jobbergate_cli.constants import SortOrder
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.render import StyleMapper, render_single_result, terminal_message
from jobbergate_cli.requests import make_request
from jobbergate_cli.schemas import ApplicationResponse, JobbergateContext
from jobbergate_cli.subapps.applications.tools import fetch_application_data, save_application_files, upload_application
from jobbergate_cli.subapps.pagination import handle_pagination


# TODO: move hidden field logic to the API
HIDDEN_FIELDS = [
    "cloned_from_id",
    "created_at",
    "is_archived",
    "template_files",
    "template_vars",
    "updated_at",
    "workflow_files",
]

ID_NOTE = """

    This id represents the primary key of the application in the database. It
    will always be a unique integer and is automatically generated by the server
    when an Application is created. All applications receive an id, so it may
    be used to target a specific instance of an application whether or not it
    is provided with a human-friendly "identifier".
"""


IDENTIFIER_NOTE = """

    The identifier allows the user to access commonly used applications with a
    friendly name that is easy to remember. Identifiers should only be used
    for applications that are frequently used or should be easy to find in the list.
    An identifier may be added, removed, or changed on an existing application.
"""


style_mapper = StyleMapper(
    application_id="green",
    name="cyan",
    identifier="magenta",
)


app = typer.Typer(help="Commands to interact with applications")


@app.command("list")
def list_all(
    ctx: typer.Context,
    show_all: bool = typer.Option(False, "--all", help="Show all applications, even the ones without identifier"),
    user_only: bool = typer.Option(False, "--user", help="Show only applications owned by the current user"),
    search: Optional[str] = typer.Option(None, help="Apply a search term to results"),
    sort_order: SortOrder = typer.Option(SortOrder.DESCENDING, help="Specify sort order"),
    sort_field: Optional[str] = typer.Option("id", help="The field by which results should be sorted"),
):
    """
    Show available applications
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx is not None
    assert jg_ctx.client is not None

    params: Dict[str, Any] = dict(
        include_null_identifier=show_all,
        user_only=user_only,
    )
    if search is not None:
        params["search"] = search
    if sort_order is not SortOrder.UNSORTED:
        params["sort_ascending"] = sort_order is SortOrder.ASCENDING
    if sort_field is not None:
        params["sort_field"] = sort_field

    handle_pagination(
        jg_ctx=jg_ctx,
        url_path="/jobbergate/job-script-templates",
        abort_message="Couldn't retrieve applications list from API",
        params=params,
        title="Applications List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=ApplicationResponse,
    )


@app.command()
def get_one(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        "--id",
        "-i",
        help=f"The specific id of the application. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
):
    """
    Get a single application by id or identifier
    """
    jg_ctx: JobbergateContext = ctx.obj
    result = fetch_application_data(jg_ctx, id=id, identifier=identifier)
    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Application",
    )


@app.command()
def create(
    ctx: typer.Context,
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="The name of the application to create",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
    application_path: pathlib.Path = typer.Option(
        ...,
        "--application-path",
        "-a",
        help="The path to the directory where the application files are located",
    ),
    application_desc: Optional[str] = typer.Option(
        None,
        help="A helpful description of the application",
    ),
):
    """
    Create a new application.
    """
    req_data = dict()
    req_data["name"] = name
    if identifier:
        req_data["identifier"] = identifier

    if application_desc:
        req_data["description"] = application_desc

    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    result = cast(
        Dict[str, Any],
        make_request(
            jg_ctx.client,
            "/jobbergate/job-script-templates",
            "POST",
            expected_status=201,
            abort_message="Request to create application was not accepted by the API",
            support=True,
            json=req_data,
        ),
    )
    application_id = result["id"]
    result["application_uploaded"] = False

    try:
        if application_path is not None:
            upload_application(jg_ctx, application_path, application_id=application_id, application_identifier=None)
            result["application_uploaded"] = True
    except Abort as e:
        raise e
    finally:
        if not result["application_uploaded"]:
            terminal_message(
                """
                The application files could not be uploaded.

                Try running the `update` command including the application path to re-upload.
                """,
                subject="File upload failed",
                color="yellow",
            )
        render_single_result(
            jg_ctx,
            result,
            hidden_fields=HIDDEN_FIELDS,
            title="Created Application",
        )


@app.command()
def update(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        "--id",
        "-i",
        help=f"The specific id of the application to update. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application to update. {IDENTIFIER_NOTE}",
    ),
    application_path: Optional[pathlib.Path] = typer.Option(
        None,
        "--application-path",
        "-a",
        help="The path to the directory where the application files are located",
    ),
    update_identifier: Optional[str] = typer.Option(
        None,
        help="Optional new application identifier to be set",
    ),
    application_desc: Optional[str] = typer.Option(
        None,
        help="Optional new application description to be set",
    ),
    application_name: Optional[str] = typer.Option(
        None,
        help="Optional new application name to be set",
    ),
):
    """
    Update an existing application.
    """
    identification: Any = id
    if id is None and identifier is None:
        terminal_message(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif id is not None and identifier is not None:
        terminal_message(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif identifier is not None:
        identification = identifier

    req_data = dict()

    if update_identifier:
        req_data["identifier"] = update_identifier

    if application_desc:
        req_data["description"] = application_desc

    if application_name:
        req_data["name"] = application_name

    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    if req_data:
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/{identification}",
            "PUT",
            expect_response=False,
            abort_message="Request to update application was not accepted by the API",
            support=True,
            json=req_data,
            expected_status=200,
        )

    try:
        if application_path is not None:
            upload_application(jg_ctx, application_path, application_id=id, application_identifier=identifier)
    except Abort as e:
        terminal_message(
            "The application files could not be uploaded.",
            subject="File upload failed",
            color="yellow",
        )
        raise e
    finally:
        if not id and update_identifier:
            # We need to fetch from new identifier if it was updated
            identifier = update_identifier
        result = fetch_application_data(jg_ctx, id=id, identifier=identifier)

        render_single_result(
            jg_ctx,
            result,
            hidden_fields=HIDDEN_FIELDS,
            title="Updated Application",
        )


@app.command()
def delete(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        "--id",
        "-i",
        help=f"The specific id of the application to delete. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application to update. {IDENTIFIER_NOTE}",
    ),
):
    """
    Delete an existing application.
    """
    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    identification: Any = id
    if id is None and identifier is None:
        terminal_message(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif id is not None and identifier is not None:
        terminal_message(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif identifier is not None:
        identification = identifier

    # Delete the upload. The API will also remove the application data files
    make_request(
        jg_ctx.client,
        f"/jobbergate/job-script-templates/{identification}",
        "DELETE",
        expected_status=204,
        expect_response=False,
        abort_message="Request to delete application was not accepted by the API",
        support=True,
    )
    terminal_message(
        """
        The application was successfully deleted.
        """,
        subject="Application delete succeeded",
    )


@app.command()
def download_files(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        help=f"The specific id of the application. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
):
    """
    Download the files from an application to the current working directory.
    """
    jg_ctx: JobbergateContext = ctx.obj

    result = fetch_application_data(jg_ctx, id=id, identifier=identifier)
    saved_files = save_application_files(
        jg_ctx,
        application_data=result,
        destination_path=pathlib.Path.cwd(),
    )
    terminal_message(
        dedent(
            """
            A total of {} application files were successfully downloaded.
            """.format(len(saved_files))
        ).strip(),
        subject="Application download succeeded",
    )


@app.command()
def clone(
    ctx: typer.Context,
    id: Optional[int] = typer.Option(
        None,
        help=f"The specific id of the application. {ID_NOTE}",
    ),
    identifier: Optional[str] = typer.Option(
        None,
        help=f"The human-friendly identifier of the application. {IDENTIFIER_NOTE}",
    ),
    application_identifier: Optional[str] = typer.Option(
        None,
        help="""
        Optional new application identifier to override the original.

        Notice this can not match an existing identifier, including the one this entry is going to be cloned from.
        """,
    ),
    application_desc: Optional[str] = typer.Option(
        None,
        help="Optional new application description to override the original",
    ),
    application_name: Optional[str] = typer.Option(
        None,
        help="Optional new application name to override the original",
    ),
):
    """
    Clone an application, so the user can own and modify a copy of it.
    """
    identification: Any = id
    if id is None and identifier is None:
        terminal_message(
            """
            You must supply either [yellow]id[/yellow] or [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif id is not None and identifier is not None:
        terminal_message(
            """
            You may not supply both [yellow]id[/yellow] and [yellow]identifier[/yellow].
            """,
            subject="Invalid params",
        )
        raise typer.Exit()
    elif identifier is not None:
        identification = identifier

    req_data = dict()

    if application_identifier:
        req_data["identifier"] = application_identifier

    if application_desc:
        req_data["description"] = application_desc

    if application_name:
        req_data["name"] = application_name

    jg_ctx: JobbergateContext = ctx.obj

    # Make static type checkers happy
    assert jg_ctx.client is not None

    result = cast(
        ApplicationResponse,
        make_request(
            jg_ctx.client,
            f"/jobbergate/job-script-templates/clone/{identification}",
            "POST",
            json=req_data,
            expected_status=201,
            abort_message=f"Couldn't clone application {identification} from API",
            response_model_cls=ApplicationResponse,
            support=True,
        ),
    )

    render_single_result(
        jg_ctx,
        result,
        hidden_fields=HIDDEN_FIELDS,
        title="Cloned Application",
    )
